@@ -1,7 +1,7 @@
 #
 # api.py
 #
-# Copyright (C) 2017-2021 frnmst (Franco Masotti) <franco.masotti@live.com>
+# Copyright (C) 2017-2021 frnmst (Franco Masotti) <franco.masotti@tutanota.com>
 #
 # This file is part of md-toc.
 #
@@ -20,496 +20,23 @@
 #
 """The main file."""
 
-import fpyutils
+import copy
 import re
 import sys
-from .exceptions import (GithubOverflowCharsLinkLabel, GithubEmptyLinkLabel,
+
+import fpyutils
+
+from . import cmark, generic
+from .constants import common_defaults
+from .constants import parser as md_parser
+from .exceptions import (CannotTreatUnicodeString, GithubEmptyLinkLabel,
+                         GithubOverflowCharsLinkLabel,
                          GithubOverflowOrderedListMarker,
                          StdinIsNotAFileToBeWritten,
-                         TocDoesNotRenderAsCoherentList, StringCannotContainNewlines,
-                         CannotTreatUnicodeString)
-from .constants import parser as md_parser, common_defaults
+                         StringCannotContainNewlines,
+                         TocDoesNotRenderAsCoherentList)
 
 
-##########################
-# cmark specific Classes #
-##########################
-class _DLLNode:
-    r"""A list node with attributes useful for processing emphasis."""
-
-    def __init__(self, delim_char: str, length: int, start_index: int):
-        self.delim_char = delim_char
-        self.length = length
-        self.offset = 0
-        self.original_length = length
-        self.active = True
-        self.start_index = start_index
-        self.previous = None
-        self.next = None
-        self.can_open = False
-        self.can_close = False
-
-    def __str__(self):
-        if self.delim_char is not None:
-            if self.previous is None:
-                previous = 'None'
-            else:
-                previous = hex(id(self.previous))
-            if self.next is None:
-                next = 'None'
-            else:
-                next = hex(id(self.next))
-
-            el = '== element ' + hex(id(self)) + " =="
-            de = 'delim_char = ' + self.delim_char
-            le = 'length = ' + str(self.length)
-            of = 'offset = ' + str(self.offset)
-            oi = 'original_length = ' + str(self.original_length)
-            ac = 'active = ' + str(self.active)
-            st = 'start_index = ' + str(self.start_index)
-            pr = 'previous = ' + previous
-            ne = 'next = ' + next
-            co = 'can_open = ' + str(self.can_open)
-            cc = 'can_close = ' + str(self.can_close)
-
-            return (el + '\n' + de + '\n' + le + '\n' + of + '\n' + oi + '\n'
-                    + ac + '\n' + st + '\n' + pr + '\n' + ne + '\n' + co + '\n'
-                    + cc + '\n')
-
-
-class _DLL:
-    r"""A double linked list useful for processing emphasis."""
-
-    def __init__(self):
-        self.start = None
-        self.end = None
-
-    def push(self, node: _DLLNode):
-        r"""Add a new node."""
-        if self.start is None and self.end is None:
-            self.start = self.end = node
-        else:
-            self.end.next = node
-            node.previous = self.end
-            self.end = self.end.next
-            node.next = None
-
-    def pop(self) -> _DLLNode:
-        r"""Remove the last node."""
-        node = None
-        if self.start is None and self.end is None:
-            pass
-        else:
-            node = self.end
-            self.end = node.previous
-
-            if self.end is None:
-                self.start = None
-
-        return node
-
-    def extract(self, node: _DLLNode):
-        r"""Remove a specific node."""
-        if node is None:
-            pass
-        elif self.start is not None and self.end is not None:
-            if node == self.end:
-                self.end = node.previous
-
-                if self.end is None:
-                    self.start = None
-            else:
-                node.next.previous = node.previous
-            if node.previous is not None:
-                node.previous.next = node.next
-            else:
-                self.start = node.next
-
-    def scroll(self):
-        r"""Print the list."""
-        x = self.start
-        while x is not None:
-            print(x)
-            x = x.next
-
-
-#####################
-# Generic functions #
-#####################
-# _ctoi and _isascii taken from cpython source Lib/curses/ascii.py
-# See:
-# https://github.com/python/cpython/blob/283de2b9c18e38c9a573526d6c398ade7dd6f8e9/Lib/curses/ascii.py#L48
-# https://github.com/python/cpython/blob/283de2b9c18e38c9a573526d6c398ade7dd6f8e9/Lib/curses/ascii.py#L56
-#
-# These 2 functions are released under the
-# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
-# See https://directory.fsf.org/wiki/License:Python-2.0.1
-def _ctoi(c: str):
-    if not len(c) == 1:
-        raise ValueError
-
-    retval = c
-    if isinstance(c, str):
-        retval = ord(c)
-
-    return retval
-
-
-def _isascii(c):
-    return 0 <= _ctoi(c) <= 127
-
-
-############################
-# cmark specific functions #
-############################
-def _cmark_is_space(char: int, parser: str = 'github') -> bool:
-    # license D applies here. See docs/markdown_specification.rst
-    value = False
-    if chr(char) in md_parser[parser]['pseudo-re']['UWC']:
-        value = True
-
-    return value
-
-
-def _cmark_is_punctuation(char: int, parser: str = 'github') -> bool:
-    # license D applies here. See docs/markdown_specification.rst
-    value = False
-    if chr(char) in md_parser[parser]['pseudo-re']['PC']:
-        value = True
-
-    return value
-
-
-def _cmark_ispunct(char: int, parser: str = 'github') -> bool:
-    # license D applies here. See docs/markdown_specification.rst
-
-    value = False
-    if chr(char) in md_parser[parser]['pseudo-re']['APC']:
-        value = True
-
-    return value
-
-
-def _cmark_utf8proc_charlen(line: str, line_length: int) -> int:
-    # license E applies here. See docs/markdown_specification.rst
-
-    if not line_length:
-        return 0
-
-    # Use length = 1 instead of the utf8proc_utf8class[256]
-    # list.
-    # For example:
-    # len('ł') == 2 # in Python 2
-    # len('ł') == 1 # in Python 3
-    # See the documentation.
-    # In Python 3 since all strings are unicode by default
-    # they all have length of 1.
-    length = 1
-    if len(line) > 1:
-        # See
-        # https://docs.python.org/3/howto/unicode.html#comparing-strings
-        raise CannotTreatUnicodeString
-
-    if not length:
-        return -1
-
-    if line_length >= 0 and length > line_length:
-        return -line_length
-
-    for i in range(1, length):
-        if (ord(line[i]) & 0xC0) != 0x80:
-            return -i
-
-    return length
-
-
-def _cmark_utf8proc_iterate(line: str, line_len: int) -> tuple:
-    # license E applies here. See docs/markdown_specification.rst
-
-    length = 0
-    uc = -1
-    dst = -1
-
-    length = _cmark_utf8proc_charlen(line, line_len)
-    if length < 0:
-        return -1, dst
-
-    if length == 1:
-        uc = ord(line[0])
-    elif length == 2:
-        uc = ((ord(line[0]) & 0x1F) << 6) + (ord(line[1]) & 0x3F)
-        if uc < 0x80:
-            uc = -1
-    elif length == 3:
-        uc = ((ord(line[0]) & 0x0F) << 12) + ((ord(line[1]) & 0x3F) << 6) + (ord(line[2]) & 0x3F)
-        if uc < 0x800 or (uc >= 0xD800 and uc < 0xE000):
-            uc = -1
-    elif length == 4:
-        uc = (((ord(line[0]) & 0x07) << 18) + ((ord(line[1]) & 0x3F) << 12) +
-              ((ord(line[2]) & 0x3F) << 6) + (ord(line[3]) & 0x3F))
-        if uc < 0x10000 or uc >= 0x110000:
-            uc = -1
-
-    if uc < 0:
-        return -1, dst
-
-    dst = uc
-
-    return length, dst
-
-
-def _cmark_scan_delims(line: str, c: str, pos: int) -> tuple:
-    # license D applies here. See docs/markdown_specification.rst
-
-    left_flanking = False
-    right_flanking = False
-    can_open = False
-    can_close = False
-    numdelims = 0
-    before_char_pos = 0
-    before_char = 0
-    after_char = 0
-
-    if pos == 0:
-        before_char = 10
-    else:
-        before_char_pos = pos - 1
-        # walk back to the beginning of the UTF_8 sequence
-        while _cmark_peek_char(line, before_char_pos) >> 6 == 2 and before_char_pos > 0:
-            before_char_pos -= 1
-
-        length, before_char = _cmark_utf8proc_iterate(line[before_char_pos:before_char_pos + 1],
-                                                      pos - before_char_pos)
-
-        if length == -1:
-            before_char = 10
-
-    if c == '\'' or c == '"':
-        numdelims += 1
-        pos += 1
-    else:
-        while chr(_cmark_peek_char(line, pos)) == c:
-            numdelims += 1
-            pos += 1
-
-    length, after_char = _cmark_utf8proc_iterate(line[pos:pos + 1], len(line) - pos)
-
-    if length == -1:
-        after_char = 10
-
-    if (numdelims > 0
-       and (not _cmark_is_space(after_char))
-       and (not _cmark_is_punctuation(after_char)
-            or _cmark_is_space(before_char)
-            or _cmark_is_punctuation(before_char))):
-        left_flanking = True
-
-    if (numdelims > 0
-       and (not _cmark_is_space(before_char))
-       and (not _cmark_is_punctuation(before_char)
-            or _cmark_is_space(after_char)
-            or _cmark_is_punctuation(after_char))):
-        right_flanking = True
-
-    if c == '_':
-        if (left_flanking
-           and (not right_flanking or _cmark_is_punctuation(before_char))):
-            can_open = True
-        if (right_flanking
-           and (not left_flanking or _cmark_is_punctuation(after_char))):
-            can_close = True
-
-    elif c == '\'' or c == '"':
-        if (left_flanking and not right_flanking and
-           before_char != ']' and before_char != ')'):
-            can_open = True
-        can_close = right_flanking
-    else:
-        can_open = left_flanking
-        can_close = right_flanking
-
-    return numdelims, can_open, can_close, pos
-
-
-def _cmark_handle_delim(delimiter_stack: _DLL, line: str, c: str, start_index: int) -> tuple:
-    # license D applies here. See docs/markdown_specification.rst
-
-    numdelims, can_open, can_close, pos = _cmark_scan_delims(line, c, start_index)
-
-    if (can_open or can_close) and (not (c == '\'' or c == '"')):
-        current = _DLLNode(c, numdelims, start_index)
-        current.can_open = can_open
-        current.can_close = can_close
-        delimiter_stack.push(current)
-
-    return numdelims, pos
-
-
-def _cmark_peek_char(line: str, pos: int) -> int:
-    # license D applies here. See docs/markdown_specification.rst
-
-    if pos < len(line):
-        return ord(line[pos])
-    else:
-        return 0
-
-
-def _cmark_remove_emph(delimiter_stack: _DLL, opener: _DLLNode, closer: _DLLNode, ignore: list):
-    # license D applies here. See docs/markdown_specification.rst
-
-    opener_num_chars = opener.length
-    closer_num_chars = closer.length
-
-    # calculate the actual number of characters used from this closer
-    if closer_num_chars >= 2 and opener_num_chars >= 2:
-        use_delims = 2
-    else:
-        use_delims = 1
-
-    # remove used characters from associated inlines.
-    opener_num_chars -= use_delims
-    closer_num_chars -= use_delims
-    opener.length = opener_num_chars
-    closer.length = closer_num_chars
-
-    # free delimiters between opener and closer
-    delim = closer.previous
-    while delim is not None and delim != opener:
-        tmp_delim = delim.previous
-        delimiter_stack.extract(delim)
-        delim = tmp_delim
-
-    opener_relative_start = opener.start_index + opener.original_length - opener.offset - use_delims
-    opener_relative_end = opener.start_index + opener.original_length - opener.offset
-
-    closer_relative_start = closer.start_index + closer.original_length - closer.offset - use_delims
-    closer_relative_end = closer.start_index + closer.original_length - closer.offset
-
-    ignore.append(range(opener_relative_start, opener_relative_end))
-    ignore.append(range(closer_relative_start, closer_relative_end))
-
-    opener.offset += use_delims
-    closer.offset += use_delims
-
-    # if opener has 0 characters, remove it and its associated inline
-    if opener_num_chars == 0:
-        delimiter_stack.extract(opener)
-
-    # if closer has 0 characters, remove it and its associated inline
-    if closer_num_chars == 0:
-        # remove closer from list
-        tmp_delim = closer.next
-        delimiter_stack.extract(closer)
-        closer = tmp_delim
-
-    return closer
-
-
-def _cmark_process_emphasis(delimiter_stack: _DLL, ignore: list) -> list:
-    # license D applies here. See docs/markdown_specification.rst
-
-    # Store indices.
-    openers_bottom = [None, None, None, None, None, None]
-
-    stack_bottom = None
-
-    # delimiter *closer = subj->last_delim;
-    closer = delimiter_stack.end
-
-    # move back to first relevant delim.
-    while closer is not None and closer.previous is not stack_bottom:
-        closer = closer.previous
-
-    # now move forward, looking for closers, and handling each
-    while closer is not None:
-        if closer.can_close:
-            if closer.delim_char == '"':
-                openers_bottom_index = 0
-            elif closer.delim_char == '\'':
-                openers_bottom_index = 1
-            elif closer.delim_char == '_':
-                openers_bottom_index = 2
-            elif closer.delim_char == '*':
-                openers_bottom_index = 3 + (closer.length % 3)
-            else:
-                raise ValueError
-
-            # Now look backwards for first matching opener:
-            opener = closer.previous
-            opener_found = False
-            while opener is not None and opener != openers_bottom[openers_bottom_index]:
-                if opener.can_open and opener.delim_char == closer.delim_char:
-                    # interior closer of size 2 can't match opener of size 1
-                    # or of size 1 can't match 2
-                    if (not (closer.can_open or opener.can_close)
-                       or closer.length % 3 == 0
-                       or (opener.length + closer.length) % 3 != 0):
-                        opener_found = True
-                        break
-                opener = opener.previous
-
-            old_closer = closer
-
-            if closer.delim_char == '*' or closer.delim_char == '_':
-                if opener_found:
-                    closer = _cmark_remove_emph(delimiter_stack, opener, closer, ignore)
-                else:
-                    closer = closer.next
-            elif closer.delim_char == '\'' or closer.delim_char == '"':
-                closer = closer.next
-            if not opener_found:
-                # set lower bound for future searches for openers
-                openers_bottom[openers_bottom_index] = old_closer.previous
-                if not old_closer.can_open:
-                    # we can remove a closer that can't be an
-                    # opener, once we've seen there's no
-                    # matching opener:
-                    delimiter_stack.extract(old_closer)
-        else:
-            closer = closer.next
-
-    # free all delimiters in list until stack_bottom:
-    while delimiter_stack.end is not None and delimiter_stack.end != stack_bottom:
-        delimiter_stack.extract(delimiter_stack.end)
-
-
-def _cmark_handle_backslash(line: str, char: str, pos: int):
-    r"""Parse backslash-escape or just a backslash, returning an inline.
-
-    .. note: license D applies here. See docs/markdown_specification.rst
-    """
-    pos += 1
-    nextchar = _cmark_peek_char(line, pos)
-
-    # only ascii symbols and newline can be escaped
-    if _cmark_ispunct(nextchar):
-        pos += 1
-
-    return pos
-
-
-def _cmark_parse_inline(delimiter_stack: _DLL, line: str, pos: int) -> tuple:
-    r"""Handle all the different elements of a string.
-
-    ..note: license D applies here. See docs/markdown_specification.rst
-    """
-    numdelim = 0
-
-    c = _cmark_peek_char(line, pos)
-    if c == 0:
-        return 0
-    elif chr(c) == '\\':
-        numdelim = _cmark_handle_backslash(line, chr(c), pos)
-    elif chr(c) == '*' or chr(c) == '_' or chr(c) == '\'' or chr(c) == '"':
-        numdelim, pos = _cmark_handle_delim(delimiter_stack, line, chr(c), pos)
-    # TODO: images, code, HTML tags detection.
-
-    return numdelim, pos
-
-
-#######
-# API #
-#######
 def write_string_on_file_between_markers(filename: str, string: str,
                                          marker: str):
     r"""Write the table of contents on a single file.
@@ -1169,24 +696,21 @@ def remove_emphasis(line: str, parser: str = 'github') -> list:
     """
     if parser in ['github', 'cmark', 'gitlab', 'commonmarker', 'goldmark', 'redcarpet']:
 
-        delimiter_stack = _DLL()
-        i = 0
-        while i < len(line):
-            numdelims, pos = _cmark_parse_inline(delimiter_stack, line, i)
+        mem = None
+        refmap = None
 
-            advance = False
-            if pos > i:
-                i = pos
-                advance = True
+        content = cmark._cmarkCmarkChunk(line, len(line), 0)
+        subj = cmark._cmark_Subject(input=line)
+        cmark._cmark_subject_from_buf(mem, 0, 0, subj, content, refmap)
 
-            if not advance:
-                i += 1
+        while not cmark._cmark_is_eof(subj):
+            cmark._cmark_parse_inline(subj)
 
         ignore = list()
-
         # When we hit the end of the input, we call the process emphasis procedure (see below), with stack_bottom = NULL.
-        _cmark_process_emphasis(delimiter_stack, ignore)
+        cmark._cmark_process_emphasis(subj, None, ignore)
         line = filter_indices_from_line(line, ignore)
+
     elif parser in ['redcarpet']:
         # TODO
         pass
@@ -1245,7 +769,7 @@ def build_anchor_link(header_text_trimmed: str,
     :raises: a built-in exception.
 
     .. note: license A applies for the redcarpet part.
-        See docs/markdown_specification.rst
+        See docs/copyright_license.rst
     """
     # Check for newlines.
     if len(replace_and_split_newlines(header_text_trimmed)) > 1:
@@ -1302,7 +826,7 @@ def build_anchor_link(header_text_trimmed: str,
                     i += 1
             # str.find() == -1 if character is not found in str.
             # https://docs.python.org/3.6/library/stdtypes.html?highlight=find#str.find
-            elif not _isascii(header_text_trimmed[i]) or STRIPPED.find(
+            elif not generic._isascii(header_text_trimmed[i]) or STRIPPED.find(
                     header_text_trimmed[i]) != -1:
                 if inserted and not stripped:
                     header_text_trimmed_middle_stage += '-'
